/*****************************************************************************
	advanet_1108_PCIVME_04.c
	2001/05/10	by M.Tadano

	cc -c -g -Wall -O -v advanet_1108_PCIVME_01.c -o advanet_1108.o \
		-I/usr/src/linux/include/ \
		-include /usr/src/linux/include/linux/modeversions.h \
		-DDDRIVER
*****************************************************************************/

#ifndef	MODULE
#define	MODULE
#endif
#ifndef	__KERNEL__
#define	__KERNEL__
#endif

#include <linux/version.h>
#include <linux/init.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/signal.h>
#include <linux/major.h>
#include <linux/module.h>
#include <linux/tqueue.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/poll.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/malloc.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#include "advanet_1108_PCIVME.h"

static unsigned int	busy = 0;
static struct pci_dev	*pci_device = NULL;
static void *pcivme_adr24, *pcivme_adr16;
static unsigned char	*kbuf = NULL;
static unsigned short	*read_buf;
static short	f_meas = OFF, int_wait = OFF;

static dDriver_t dDriver;
static vme_fn	*set_vme;

static int	major = 0;
MODULE_PARM (major, "i");

/* interrupt ******************************************************************/
static void advanet_1108_interrupt (int irq, void *dev_id, struct pt_regs *regs)
{
int	flag_int;

	flag_int = inw (PISR);
if (flag_int != 0)	printk ("Flag_int = %d/", flag_int);
	if ( 0 < (flag_int & 0xff00))	{	/* VME irq */
printk ("1108_interrupt/ VME interrupt/ ");
		flag_int = (flag_int >> 2) & 0xff;
		if (0 < (flag_int = (flag_int >> 1) & 0x1))	{ /* FEINT */
			;
		}
		if (0 < (flag_int = (flag_int >> 1) & 0x1))	{ /* FFINT */
			;
		}
		if (0 < (flag_int = (flag_int >> 1) & 0x1))	{ /* AFINT */
			;
		}
		if (0 < (flag_int = (flag_int >> 1) & 0x1))	{ /* SFINT */
			;
		}
		if (0 < (flag_int = (flag_int >> 1) & 0x1))	{ /* SRINT */
			;
		}
		if (0 < (flag_int = (flag_int >> 1) & 0x1))	{ /* BEINT */
			;
		}
		if (0 < (flag_int = (flag_int >> 1) & 0x1))	{ /* IRQ1 */
			;
		}
		if (0 < (flag_int = (flag_int >> 1) & 0x1))	{ /* IRQ2 */
			;
		}
		if (0 < (flag_int = (flag_int >> 1) & 0x1))	{ /* IRQ3 */
			;
		}
		if (0 < (flag_int = (flag_int >> 1) & 0x1))	{ /* IRQ4 */
			;
		}
		if (0 < (flag_int = (flag_int >> 1) & 0x1))	{ /* IRQ5 */
			;
		}
		if (0 < (flag_int = (flag_int >> 1) & 0x1))	{ /* IRQ6 */
			;
		}
		if (0 < (flag_int = (flag_int >> 1) & 0x1))	{ /* IRQ7 */
			;
		}
/*		if (int_wait == ON)	{
			wake_up_interruptible (&dDriver.queue);
		}
*/
	}
	return;
}

/* open **********************************************************************/
static int advanet_1108_open (struct inode *inode, struct file *file)
{
int	minor, result;

	printk ("1108_open\n");
	if (busy != 0) 	{
		return - EBUSY;
	}
	minor = MINOR (inode->i_rdev);
	if (minor < 0)	{
  		return -ENODEV;
	}
	busy = 1;
	MOD_INC_USE_COUNT;
	int_wait = OFF;

	dDriver.minor = minor;
	if (result = (request_irq (dDriver.irq, advanet_1108_interrupt,
		SA_INTERRUPT | SA_SHIRQ, DEVICE_NAME, pci_device)))	{
		printk ("1108_open: request_irq error: error = %d\n", 
								result);
		unregister_chrdev (dDriver.major, DEVICE_NAME);
		return result;
	}
	printk ("1108_open/ return\n");
	return 0;
}

/* close *********************************************************************/
static int advanet_1108_close (struct inode *inode, struct file *file)
{
	printk ("1108_close\n");

	f_meas = OFF;
	if (kbuf != NULL)
		kfree (kbuf);
	free_irq(dDriver.irq, pci_device);
	busy = 0;
	MOD_DEC_USE_COUNT;
	
	printk ("1108_close/ return\n");
	return 0;
}

/* read **********************************************************************/
static ssize_t advanet_1108_read 
	(struct file *file, char *buf, size_t count, loff_t *ppos)
{
int	len, i;
static unsigned long	flags;
short	*vme_adr = NULL;

	len = 0;
	set_vme = (vme_fn *)buf;
	vme_adr = pcivme_adr24 + set_vme->address;
	switch (set_vme->command)	{
	case	VME_READ_A16_D8:
		(unsigned char *)read_buf = (unsigned char *)kbuf;
		for (i = 0; i < count; i++)
			*(unsigned char *)(read_buf +i) =
				 *(unsigned char *)(vme_adr +i);
		len = count;
		break;
	case	VME_READ_A24_D16:
		read_buf = (short *)kbuf;
		for (i = 0; i < count; i++)
			*(short *)(read_buf +i) = *(short *)(vme_adr +i);
		len = count*2;
		break;
	defaults:
		read_buf = (unsigned short *)kbuf;
		break;
	}
	save_flags (flags);
	cli();
	if (copy_to_user (buf, kbuf, len) != 0)	{
		restore_flags (flags);
		printk ("1108_read/ copy_to_user ERROR !!!!\n");
		return -EFAULT;
	}
	restore_flags (flags);

	return CHANNEL_MAX*2;
}

/* write ********************************************************************/
static ssize_t advanet_1108_write (
	struct file *file, const char *buf, size_t count, loff_t *ppos)
{
static unsigned long	flags;
long	*vme_adr = NULL;
int	i, sizelen, retval;

	sizelen = (sizeof(short)+sizeof(long)+sizeof(int) + 2*sizeof(char)) 
		* sizeof(char);
	save_flags (flags);
	cli();
	if ((retval = copy_from_user (kbuf, buf, count + sizelen)) != 0)	{
		printk ("1108_write/ copy_from_user ERROR (retval=%d)!!!!\n",
					retval);
		return -EFAULT;
	}
	set_vme = (vme_fn *)kbuf;
	switch	(set_vme->command)	{
	case	VME_WRITE_A16_D8:
		vme_adr = pcivme_adr16 + set_vme->address;
		for ( i = 0; i < count; i++)
			*(unsigned char *)(vme_adr +i) =
					*(unsigned char *)(kbuf +sizelen +i);
		*vme_adr = set_vme->w_bit;
		break;
	case	VME_WRITE_A24_D16:
		sizelen /= sizeof(short);
		vme_adr = pcivme_adr24 + set_vme->address;
		for ( i = 0; i < count; i++)
			*(short *)(vme_adr +i) = *(short *)(kbuf + sizelen +i);
		*vme_adr = set_vme->w_bit;
		break;
	defaults:
		return 0;
		break;
	}

	return 0;
}

/* ioctl *********************************************************************/
static int advanet_1108_ioctl(struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg)
{
short	*vme_adr = NULL;
int	i, len;
static unsigned long	flags;

	len = 0;
	set_vme = (vme_fn *)arg;
	read_buf = (unsigned short *)kbuf;

	switch (cmd)	{
	case	VME_READ_A16_D8:
		vme_adr = pcivme_adr16 + set_vme->address;
		for ( i = 0; i < set_vme->count; i++)
			*(unsigned char *)(read_buf +i) =
				*(unsigned char *)(vme_adr +i);
		len = set_vme->count;
		break;
	case	VME_READ_A24_D16:
		vme_adr = pcivme_adr24 + set_vme->address;
		for (i = 0; i < set_vme->count; i++)
			*(short *)(read_buf +i) = *(short *)(vme_adr +i);
		len = set_vme->count *2;
		break;
	case	VME_WRITE_A16_D8:
		vme_adr = pcivme_adr16 + set_vme->address;
		for (i = 0; i < set_vme->count; i++)	
			*(unsigned char *)(vme_adr +i) =
				(unsigned char)set_vme->data[i];
		break;
	case	VME_WRITE_A24_D16:
		vme_adr = pcivme_adr24 + set_vme->address;
		for (i = 0; i < set_vme->count; i++)	
			*(short *)(vme_adr +i) = (short)set_vme->data[i];
		break;
	case	LOSS_READ:
		for ( i = 0; i < CHANNEL_MAX*2; i++)
			*(short *)(read_buf +i) = *(short *)(vme_adr +i);
		len = CHANNEL_MAX*4;
		break;
	defaults:
		return 0;
		break;
	}
	if ( READ_COMMAND <= cmd && cmd < WRITE_COMMAND)	{
		save_flags (flags);
		cli();
		if (copy_to_user ((char *)arg, (char *)kbuf, len) != 0)	{
			restore_flags (flags);
			printk ("1108_read/ copy_to_user ERROR !!!!\n");
			return -EFAULT;
		}
		restore_flags (flags);
	}
	return 0;
}

/* file operation ************************************************************/
static struct file_operations advanet_1108_fops =
{
  THIS_MODULE,			/* struct module *owner;*/
  NULL,				/* advanet_1108_llseek */
  advanet_1108_read,		/* advanet_1108_read */
  advanet_1108_write,		/* advanet_1108_write */
  NULL,				/* advanet_1108_readdir */
  NULL,				/* advanet_1108_poll */
  advanet_1108_ioctl,		/* advanet_1108_ioctl */
  NULL,				/* advanet_1108_mmap */
  advanet_1108_open,		/* advanet_1108_open */
  NULL,				/* advanet_1108_flush */
  advanet_1108_close,		/* advanet_1108_release */
  NULL,				/* advanet_1108_fsync */
  NULL,				/* advanet_1108_fasync */
  NULL,				/* advanet_1108_lock */
  NULL,				/* advanet_1108_readv */
  NULL				/* advanet_1108_writev */
};

/* init **********************************************************************/
int init_module (void)
{
struct pci_dev	*ret_dev;
int	result;

printk ("init_module\n");
	ret_dev = pci_find_device (ADVANET_1108_PCIVME_BOARD_VENDOR,
                              ADVANET_1108_PCIVME_BOARD_DEVICE, pci_device);
	if (!ret_dev)
		return -ENODEV;         /* not found */
	pci_device = ret_dev;
	result = pci_enable_device (pci_device);
	if (result)
		return result;
	dDriver.iobase0 = pci_resource_start (pci_device, 0);
	dDriver.iobase1 = pci_resource_start (pci_device, 1);
	dDriver.iobase2 = pci_resource_start (pci_device, 2);
	dDriver.iobase3 = pci_resource_start (pci_device, 3);
	dDriver.iobase4 = pci_resource_start (pci_device, 4);
	dDriver.iobase5 = pci_resource_start (pci_device, 5);
	dDriver.irq = pci_device ->irq;
	result = register_chrdev (major, DEVICE_NAME, &advanet_1108_fops);
	if (result < 0)	{
		unregister_chrdev (dDriver.major, DEVICE_NAME);
		return - EIO;
	}
	if ((pcivme_adr24 = (void *)ioremap (
			dDriver.iobase0, 0x1000000-0x10000)) == NULL)	{
		printk ("init_module(1108)/ ioremap error(_adr24) !!!!!\n");
		unregister_chrdev (dDriver.major, DEVICE_NAME);
		return 0;
	}
	if ((pcivme_adr16 = (void *)ioremap (
			dDriver.iobase0+0xff0000, 0x10000)) == NULL)	{
		printk ("init_module(1108)/ ioremap error(_adr16) !!!!!\n");
		unregister_chrdev (dDriver.major, DEVICE_NAME);
		return 0;
	}

	if (major == 0)
	  	dDriver.major = result;
	else
		dDriver.major = major;
	init_waitqueue_head (&dDriver.queue);		/* add */
	kbuf = kmalloc (0x200, GFP_KERNEL);

printk ("iobase0=%x ", dDriver.iobase0);
printk ("iobase1=%x ", dDriver.iobase1);
printk ("iobase2=%x ", dDriver.iobase2);
printk ("iobase3=%x ", dDriver.iobase3);
printk ("iobase4=%x ", dDriver.iobase4);
printk ("iobase5=%x ", dDriver.iobase5);
printk ("irq=%d \n", dDriver.irq);

printk ("init_module/ 1108/ end\n");
	return 0;
}

/* cleanup *******************************************************************/
void cleanup_module (void)
{
printk ("cleanup_module\n");
	release_region (dDriver.iobase0, 64);
	if (kbuf != NULL)
		kfree (kbuf);
	vfree (pcivme_adr24);
	vfree (pcivme_adr16);
	unregister_chrdev (dDriver.major, DEVICE_NAME);
printk ("cleanup_module/ end\n");
}
/* end of file ***************************************************************/
